# -*- coding: utf-8 -*-
"""
Created on Tue Mar 31 13:30:57 2020

@author: simsh
"""

#!/usr/bin/python3


import pandas as pd

import rtasConstants as constants

from sklearn.metrics import r2_score
import numpy as np
import matplotlib.pyplot as plt 


#this is the key
#called by plot_util_quad_from_file
def add_plot(numSubtasks, prob, utilDist, smtDist, maxSol, maxDist, numLayers, plotData, scenarioString):
    

    #add x-axis tick marks to graph
    '''
    minTotalUtil=0
    maxTotalUtil=coreCount*2
    ticks=[]
    for j in range(minTotalUtil, maxTotalUtil+1):
        ticks.append(j)
    plt.set_xticks(ticks)
    '''

    #Graph Title
    if utilDist==constants.NARROW:
        utilString="Narrow"
    elif utilDist==constants.WIDE:
        utilString="Wide"
    elif utilDist==constants.MEDIUM:
        utilString = "Medium"
    else: utilString="utilDistUnknown"
        
    if smtDist==constants.OPTIMIST:
        smtString="Optimist"
    elif smtDist==constants.OK:
        smtString="Mid-Range"
    elif smtDist==constants.PESSIMIST:
        smtString = "Pessimist"
    else: smtString="smtDistUnknown"
    
    if numLayers==0:
        probString="P=" + str(prob)
    else:
        probString=str(numLayers) + " Layers, P=" + str(prob)
    
    #mean relative util
    meanRU=plotData['smtUtil'].sum()/plotData['baseUtil'].sum()
    formatRU="{:.2f}".format(meanRU)
    
    #mean relative CC
    meanRCC=plotData['smtCores'].sum()/plotData['baseCores'].sum()
    formatRCC="{:.2f}".format(meanRCC)
    
    #mean time
    meanTime=plotData['solverTime'].mean()
    intTime=int(meanTime)
    
    coreReduceFreq=plotData['coresReduced'].sum()/plotData.shape[0]
    formatCoreReduce="{:.2f}".format(coreReduceFreq)
    
    titleString=str(numSubtasks) + " Subtasks, " + \
           probString + ", " + \
           utilString + ", " + \
           smtString + ", " + \
           "K=" + str(maxDist) + "\n" + \
           "(" + str(formatRCC) + ", " +\
           str(formatRU) + ", " +\
           str(formatCoreReduce) + ", " +\
           str(intTime) +")"# + "\n" +\
           #"reference lines: y=0.5x, y=x, y=1.5x"
    
    fileName=str(numSubtasks) + "-" + str(numLayers) + "-" + str(prob) + "-" + \
        utilString + "-" + smtString + "-" + str(maxDist)
           

    #plot the results
    
    #change in Util
    #color = purple
    newGraph=plotData.plot(kind='scatter', x='baseUtil', y='smtUtil',
    color='#7570b3', label='SMT Util.', marker="x")
    #trendline for util


    
    #newGraph.set_ylim(0.45, 1.01)
    
    #smt Cores
    #color green
    plotData.plot(kind='scatter', x='baseUtil', y='smtCores',
    color='#1b9e77', label='SMT Cores', ax=newGraph, title=titleString, marker="D")
    
    #base core
    #color orange
    plotData.plot(kind='scatter', x='baseUtil', y='baseCores',
                  color='#d95f02', label='Base Cores', ax=newGraph, marker='.')
    
    #reference line
    minUtil=plotData['baseUtil'].min()
    maxUtil=plotData['baseUtil'].max()

    newGraph.set_xlabel('Base Util.')
    newGraph.set_ylabel('SMT Util, Cores')
    newGraph.figure.savefig("results/03-14/singleDAG/" + fileName + "-new.pdf")                    
                           


def summarize(scenarioString, scenarioDF, outFile):
    #for each of the values we want to summarize
    scenarioString=scenarioString + ',' + str(scenarioDF.shape[0])
    for colName in ['smtUtil/base', 'smtCores/base', 'coreReduceFreq', 'solverTime', 'optimalGap']:
        if not colName == 'coreReduceFreq':
            column=scenarioDF[colName]
            #scenarioString.join(',' + str(column.min), str(column.mean), str(column.max))
            scenarioString=(scenarioString + 
            ',' + str(column.min()) +
            ',' + str(column.max())
            )
        if colName == 'smtUtil/base':
            mean=scenarioDF['smtUtil'].sum()/scenarioDF['baseUtil'].sum()
            scenarioString=scenarioString + ',' + str(mean)
        elif colName == 'smtCores/base':
            mean=scenarioDF['smtCores'].sum()/scenarioDF['baseCores'].sum()
            scenarioString=scenarioString + ',' + str(mean)
        elif colName == 'coreReduceFreq':
            # how often is the core count reduced?
            totalCount=scenarioDF.shape[0]
            reductionCount=scenarioDF['coresReduced'].sum()
            scenarioString=scenarioString + ',' + str(reductionCount/totalCount)
            
        else:
            scenarioString=scenarioString + ',' + str(column.mean())
    with open(outFile, 'a') as f:
        print(scenarioString, file=f)
    
    


#Get data

inFile='results/10-29/combined.csv'
allData=pd.read_csv(inFile)

summaryFile="results/03-14/singleDAG/summary.csv"

'''
numSubtasks=40
prob=0.5
utilDist=0
smtDist=0
maxSol=10
maxDist=10
xAxis=""
fileOut=""
'''

#make data numeric
allData['baseUtil']=pd.to_numeric(allData['baseUtil'])
allData['smtUtil']=pd.to_numeric(allData['smtUtil'])
allData['baseCores']=pd.to_numeric(allData['baseCores'])
allData['smtCores']=pd.to_numeric(allData['smtCores'])
allData['length']=pd.to_numeric(allData['length'])
allData['deadline']=pd.to_numeric(allData['deadline'])
allData['smtUtil/base']=pd.to_numeric(allData['smtUtil/base'])
allData['smtCores/base']=pd.to_numeric(allData['smtCores/base'])
allData['solverTime']=pd.to_numeric(allData['solverTime'])
allData['optimalGap']=pd.to_numeric(allData['optimalGap'])


#add in column for length/deadline
dOverL=[]
for k in range(allData.shape[0]):
    dOverL.append(allData['deadline'].iloc[k]/allData['length'].iloc[k])
    
newColumn=pd.Series(dOverL)
allData.insert(loc=0, column="dOverL", value=newColumn)
    
#Scenario defined by utilDist, smtDist, prob, nodeCount, maxDist, maxSol
#Scenario summary


#print headers to summary file

'''
with open (summaryFile, "a") as f:
    print("utilDist", "smtDist", "numLayers", "prob", "subtaskCount", "maxDist", "maxSol",
          "numTrials", 
          "utilReducMin", "utilReducMax", "utilReducMean",
          "coreReducMin", "coreReducMax", "coreReducMean", "coreReduceFreq",
          "solverTimeMin", "solverTimeMax", "solverTimeMean",
          "optimalMin", 'optimalMax', 'optimalMean',
          file=f, sep=',')
'''

for u in [0, 1, 2]:
    for s in [0, 1, 2]:
        for p in [0.1, 0.3, 0.5, 0.7, 0.9]:
            for c in [10, 20, 30, 40, 50, 60, 70, 80, 90, 100, 250, 500]:
                for maxDist in [1, 10, 20, 40, 80, 1000]:
                    for maxSol in [10]:
                        for numLayers in [0, 2, 4, 8, 16, 32, 64]:
                            scenarioDF= allData[(allData['utilDist']==u) &
                                                (allData['smtDist']==s) &
                                                (allData['Prob']==p) &
                                                (allData['nodeCount']==c) &
                                                (allData['maxSol']==maxSol) &
                                                (allData['maxDist']==maxDist) &
                                                (allData['numLayers']==numLayers)
                                                ]
                        
                            if (scenarioDF.shape[0>0]):
                                scenarioList=[str(u), str(s), str(numLayers), str(p), str(c), 
                                              str(maxDist), str(maxSol)]
                                scenarioString=','.join(scenarioList)
                                
                                # reset indices so they can be used moving forward
                                scenarioDF.reset_index(drop=True, inplace=True)
                                
                                #maxDist>c is an artifact of how we iterated through scenarios.
                                #has no effect, better to just say maxDist=nodeCount
                                effectiveMaxDist=min(maxDist, c)
                                
                                #new: add new column to show whether cores were decreased
                                coresReducedList=[]
                                for i in range(scenarioDF.shape[0]):
                                    if (scenarioDF.at[i, 'smtCores']<scenarioDF.at[i, 'baseCores']):
                                        coresReducedList.append(1)
                                    else:
                                        coresReducedList.append(0)
                                scenarioDF=scenarioDF.assign(coresReduced=coresReducedList)
                                    
                                
                                # cancel SMT for when it would increase the core count
                                for i in range(scenarioDF.shape[0]):
                                    if (scenarioDF.at[i, 'smtCores']>scenarioDF.at[i, 'baseCores']):
                                        scenarioDF.at[i, 'smtCores']=scenarioDF.at[i, 'baseCores']
                                        scenarioDF.at[i, 'smtUtil']=scenarioDF.at[i, 'baseUtil']
                                        scenarioDF.at[i, 'smtUtil/base']=1
                                        scenarioDF.at[i, 'smtCores/base']=1

                                        

                                summarize(scenarioString, scenarioDF, summaryFile)
                                #add_plot(c, p, u, s, maxSol, effectiveMaxDist, numLayers, scenarioDF, scenarioString)



