# -*- coding: utf-8 -*-
"""
Created on Thu Oct 17 08:53:41 2019

@author: shosborn
"""


# need to have a fully defined DAG before this can run.

from gurobipy import *
#from random import random, gauss, uniform, choice
import pandas as pd

class schedDAG:
    #create variables for frame sizes
    
    def __init__(self, dag):
        self.solver=Model("qcp")
        self.dag=dag
        
    #print("Testing if arguments work.")
    #print(taskSystem.allTasks[1].allCosts[5])
    
    
    def setSolverParams(self):
        #taskSystem=self.taskSystem
        self.solver.setParam("TimeLimit", self.dag.timeout)
        self.solver.setParam("SolutionLimit", self.dag.solutionLimit)
        self.solver.setParam(GRB.Param.Threads, self.dag.threadsPerTest)
        #For fastest peformance, set lb=ub=1
        #coreUB=(self.dag.totalCost/self.dag.deadline)*2 + 1
        self.varCoreCount=self.solver.addVar(lb=coreLB, ub=coreUB, vtype=GRB.INTEGER)
        self.solver.setObjective(self.varCoreCount, GRB.MINIMIZE)
    
    def schedule(self):
        self.setSolverParams()
        #self.createFrameSizeVars()
        self.createSchedVars()
        self.solver.optimize()
        return self.solver.getAttr(GRB.Attr.Status)
        #self.printFrameSizes()
        #print("Effective util:")
        #print(self.effectiveUtil())

    def createSchedVars(self):
        taskSystem=self.dag
        deadline=taskSystem.deadline

        self.schedVars={'taskID_1'       : [],
                        'taskID_2'       : [],
                        'instant'           : [],
                        'schedVar'          : []
                }



        for i in range(self.dag.nTotal):
            subTask1=taskSystem.allTasks[i]
           
            # constraint (ii): all subtasks scheduled
            exprScheduled = LinExpr()
            
            #constraint 4: precedence constraints
            #finishVar=self.solver.addVar(lb=1, ub=deadline, vtype=GRB.INTEGER)
            #exprFinish=linExpr()
            
            for j in range(taskSystem.nTotal):
                subTask2=taskSystem.allTasks[j]
                
                # create a variable iff there's no precedence relationship
                # need to create noPred method, min and maxStart properties

                #CONSTRAINT (i): if i and j are together, they must be together with enough time for both to finish
                exprAllOrNone=LinExpr()
                
                for t in range(0, deadline):                 
#               for t = max(subTask1.minStart, subtask2.minStart) to min(task1.maxStart, task2.maxStart):
                    self.schedVars['taskID_1'].append(i)
                    self.schedVars['taskID_2'].append(j)
                    self.schedVars['instant'].append(t)
                    
                    var = self.solver.addVar(lb=0, ub=1, vtype=GRB.BINARY)
                    self.schedVars['schedVar'].append(var)
                    
                    
                    exprAllOrNone += var
                    exprScheduled += var
                #end of t loop
                binAllOrNone=self.solver.addVar(lb=0, ub=1, vtype=GRB.BINARY)
                self.solver.addConstr(lhs=exprAllOrNone, rhs=binAllOrNone * subTask1.allCosts[j],sense=GRB.EQUAL)
                
                # end of definining constraint (i)
            # end of j loop
            
            # constraint (ii) all tasks scheduled
            self.solver.addConstr(lhs=exprScheduled, rhs=1, sense=GRB.GREATER_EQUAL)
            #end of defining constraint 2
            
            #constraint 4: precedence constraints
            #self.solver.addConstr(lhs=finishVar, rhs=exprFinish + 1, sense=GRB.GREATER_EQUAL)
            
            
        # end of i loop
        
        #convert the variables I just put together to a pandas dataframe        
        self.schedVarsP=pd.DataFrame(self.schedVars)
       
        #Constraint 3: number of items scheduled never exceeds the core count
        for t in range(0, deadline):
            exprCoreCount = LinExpr()
            
            schedVarsT=self.schedVarsP[(self.schedVarsP['instant']==t)]
            
            for i in range(taskSystem.nTotal):
                schedVars_TI=schedVarsT[(schedVarsT['taskID_1']==i)]
                subTask1=taskSystem.allTasks[i]
                for j in range(taskSystem.nTotal):
                    
                    subTask2=taskSystem.allTasks[j]
                    if i==j:
                    
                    #if (i==j or  
                    #subTask1.allCosts[j] > subTask2.allCosts[i] or 
                    #(subTask1.allCosts[j] == subTask2.allCosts[i] and i < j)):
                        # How do I find the correct variable? Can I do it without pandas?
                        '''
                        var I need is the entry in schedVars with the index that
                        matches all 3 of i, j, and time
                        '''
                        var=schedVars_TI[(schedVars_TI['taskID_2']==j)]['schedVar'].iloc[0]

                        exprCoreCount +=var
                #end of j loop
            #end of i loop
            
            self.solver.addConstr(lhs=exprCoreCount, rhs=self.varCoreCount, sense=GRB.LESS_EQUAL)
        #end of t loop
        #end of constraint # 3
 
